'''
Created on August 02, 2012

@author: William Schilthuis
email:   WilliamGS@gmail.com
website: http://williamgs.com
license: http://www.gnu.org/licenses/gpl-3.0.html

    Copyright (C) 2012 William Schilthuis

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

BRIEF
    Automatic and customizable Linear Workflow in Maya.

requires: mayapy

Change Log: {Major}.{Feature}.{Bug}
    2.4.0 : 2013-01-08
        - replaced pre-render MEL extravaganza with python scriptNode.
    2.3.0 : 2012-10-11
        - Exclusions UI (beta)
    2.2.0 : 2012-09-xx
        - Main UI
        - pickling
    2.1.0 : 2012-08-23
        - fixed exclusions, added source/destination specificity
        - added default profile functions
    2.0.0 : 2012-08-22
        - Python/pymel version.
        - primary MEL replicated. essential core module working.
        - lots of cleanup and small features yet to do.
    1.0.0
        - original MEL.
        
    BUGS:
        - Alpha is not perfect, although generally irrelevant
    TODO:
        - separate I/O LUT in preferences
        - rigorously document
        - rigorously debug
        - check with anim/expressions
        - check without MR loaded
        - save UI state to prefs (core already saved)
        - exclusion presets
        - exclusion options for remote rendering (load saved preset)
        - exclusion options for EXR checking
    
DESCRIPTION
    There are several problems which this tool aims to solve:
    1 - Maya currently only color manages file textures, and not solid colors
    2 - Maya's color management doesn't work properly.
    3 - correcting shading networks manually is tedious
    
    To calculate and output linear images, shading networks must receive linear
    data. We must manually convert all input colors to be in linear color space
    before feeding them to the shader. For shaders with simple color and/or
    texture inputs, the process is straightforward.
    
    For procedural networks and with mixed inputs (linear and non-linear) the
    process is fundamentally flawed (because maya fails to handle all colors).
    
    There are several potential strategies for linearizing networks:
    1 - linearize the color output of each individual node
    2 - carefully construct networks to linearize colors before blending
    3 - linearize the end result (just before shader input)
    Method (1) fails for inline networks (ramp1.outColor -> ramp2.colorGain) 
    Method (2) is the only mathematically correct solution, but cannot be done
        automatically and may be impossible when mixing linear and sRGB inputs.
    Method (3) is how this tool works. It is not 100% mathematically correct,
        but provides a visually expected result and is possible to automate.
        Working in the hypershade is reasonably accurate. There is still a
        problem when linear textures are not directly attached to shaders. In
        such an event, the user needs to decide how to construct the network:
        A - inversely correct linear textures, and linearize overall at end.
        B - Method (2) above
        C - avoid mixing linear and non-linear nodes in one shader input.
        In any event, this tool provides help for the process.
    For more information see: http://williamgs.com/blog/?p=139
    
ADDITIONAL NOTES
    
Features:
    1 - Prerender/Postrender scripts (including installation)
    2 - options for exclusions by node type/name/connection and src/dst/either.
    3 - manual correction tools

USAGE:
    There are two primary classes:
        gammaCorrect - contains the core modules, no GUI required. 
        UI - the graphical user interface.
    
    # GUI mode
        import wLinearWorkflow as wLW
        reload(wLW)
        wLW.loadUI()
    
    # command-line cleanup:
        import wLinearWorkflow as wLW
        reload(wLW)
        # install
        wLW.setup(True)
        # uninstall
        wLW.setup(False)
    
    # core/testing guide:
        import wLinearWorkflow as wLW
        reload(wLW)
        gc = wLW.gammaCorrect()
        gc.verbose = 3 # 1:errors 2: warn 3: info 4: debug
        gc.insertTargetgNodes()
        gc.removeTargetgNodes()
        gc.preRenderCall()
        gc.postRenderCall()
        del gc
'''

from functools import partial
import pickle
import os
import maya.cmds as mc
import maya.mel  as mm

def setup(active=True):
    '''
    Installs pre/post-render script as scriptNode. The only MEL needed is for 
    calling the scriptNode upon pre/post render.
    '''
    wLW=''
    if not len(mc.ls('wLinearWorkflow',type='script',recursive=True)):
        wLWImport = 'import wLinearWorkflow as wLW'
        wLW = mc.scriptNode(name='wLinearWorkflow',sourceType='python',
            scriptType=0, # on demand
            beforeScript='%s\nwLW.preRender()' % wLWImport,
            afterScript= '%s\nwLW.postRender()' % wLWImport)
    else:
        wLW = mc.ls('wLinearWorkflow',type='script',recursive=True)[0]
    prr = ''
    psr = ''
    if active:
        prr = 'scriptNode -eb  "%s"' % wLW
        psr = 'scriptNode -ea  "%s"' % wLW
    mc.setAttr('defaultRenderGlobals.preMel', prr,type='string')
    mc.setAttr('defaultRenderGlobals.postMel',psr,type='string')
        
    # print MEL for command line panel visibility
    mm.eval('print("// wLinearWorkflow pre/post-render [{0}] //\\n")'.format(active))
def preRender():
    '''
    PreRender MEL calls this function.
    '''
    mc.renderThumbnailUpdate(False)
    gc = gammaCorrect()
    gc.verbose = 2
    gc.target = 'all'
    #gc.exclusions = loadExclusion() 
    gc.preRenderCall()
    del gc
def postRender():
    '''
    PostRender MEL calls this function.
    '''
    gc = gammaCorrect()
    gc.verbose = 2
    gc.target = 'all'
    #gc.exclusions = loadExclusion() 
    gc.postRenderCall()
    del gc
    mc.renderThumbnailUpdate(True)
def loadUI():
    """
    Launches an instance of tool GUI.
    same as main wLinearWorkflow() MEL procedure.
    """
    LWUI = UI()
    del LWUI
def defaultExclusions(excl=1):
    '''
    Returns hard-coded exclusion presets. Used by main gammaCorrect class to
    load default exclusions.
    
    This format can be used to hard-code new presets in absence of a great UI
    '''
    if excl == 1: # default
        return {
    'type' : [
        # MATERIALS
        ('surfaceShader',           '*',                    'dst'),
        ('*',                       'transparency',         'srcdst'),
        ('*',                       'outTransparency',      'dst'),
        ('*',                       'outMatteOpacity',      'srcdst'),
        ('*',                       'miSpecularColor',      'dst'),
        ('*',                       'miColor',              'dst'),
        ('*',                       'vrEdgeColor',          'dst'),
        ('mia_material_x_passes',   'refr_color',           'dst'),
        ('mia_material_x_passes',   'refr_falloff_color',   'dst'),
        ('mia_material_x_passes',   'additional_color',     'dst'),
        # GENERIC SHADERS
        ]}
    elif excl == 2: # demo/testing preset
        return {
    'type' : [
        # MATERIALS
        ('surfaceShader',           '*',                    'dst'),
        ('*',                       'transparency',         'srcdst'),
        ('*',                       'outTransparency',      'dst'),
        ('*',                       'outMatteOpacity',      'srcdst'),
        ('*',                       'miSpecularColor',      'dst'),
        ('*',                       'miColor',              'dst'),
        ('*',                       'vrEdgeColor',          'dst'),
        ('mia_material_x_passes',   'refr_color',           'dst'),
        #('*',                       'instObjGroups',        'i'),
        #('*',                       'instObjGroups.objectGroups','i'),
        #('*',                       'objectGroups',         'i'),
        # GENERIC SHADERS
        ('envShpere',               '*',                    'src'),
        ('layeredTexture',          'inputs[1].color',      'dst')],
    'name' : [
        # MATERIALS
        ('lambert1',                '*',                    'dst'),
        ('shaderGlow1',             '*',                    'dst'),
        ('particleCloud1',          'color',                'dst'),
        # GENERIC SHADERS
        ('layeredTexture1',         'inputs[0].color',      'dst'),
        ('layeredTexture2',         'outColor',             'src')]}
    else:
        return None
def defaultProfile(prfl=1):
    '''
    Profiles are used to describe how non-linear inputs are converted to linear
    data.
    "gNodeType" describes the node type used for correction. The first item is
        the node type created for new corrections. The following items in the
        list (if any) are alternate nodes which will also be recognized, which
        allows cleanup and correction of networks constructed differently
        without needing to switch profiles.
    "gAttrs" describes the input and output attribute names for the gNodeType.
    "gSetting" describes the attribute and setting which determine the profile,
        or lookup and reverse lookup tables.
        In the case of mip_gamma_gain this is
            ["gamma",2.2 (display),0.4545 (inverse)].
        In the case of a theoretical color profile node we might see
            ["LUT","rec709","sRGB"]
        This way we can use rec709 (or gamma 1.7) for display and still assume
        our textures and inputs were saved with sRGB (or gamma 2.2) color space.
    TODO:
        - It may be more bulletproof to include a convenient way to switch
            profiles via the UI and (possibly) not include alternate gNodeTypes.
            This would probably only need to happen once: when opening a shading
            network that was created with a different type of gNodeType.
    '''
    if prfl == 1:
        return {
            'gNodeType' : ['mip_gamma_gain','gammaCorrect'],
            'gAttrs'    : [['input','inputA'],['outValue','outValueA']],
            'gSetting'  : [['gamma'],(2.2,1.0/2.2)]}
    else:
        return {
            'gNodeType' : ['gammaCorrect','mip_gamma_gain'],
            'gAttrs'    : [['value'],['outValue']],
            'gSetting'  : [['gammaX','gammaY','gammaZ'],(2.2,1.0/2.2)]}

class gammaCorrect(object):
    '''
    Core Linear Workflow object.
    1 - Create the object
    2 - (optional) set profile (which correction node and gamma/LUT settings)
    3 - (optional) set exclusions (such as Transparency, or surfaceShaders)
    4 - (optional) set verbosity
    5 - run methods at will
    
    Why ObjectOriented instead of Function programming?
    I'm using a custom class here instead of a regular functions in order to
    avoid passing massive lists of arguments to and from each command. Data such
    as exclusions, profile settings, and target mode are instead saved as
    attributes on the object and reused.

    TODO:
        - insert as source or destination? (low priority)
        - optimize: check if attributes are disabled (precheck auto exclusion?)
        - disable color management
        - regular expression on exclusion methods to handle wildcards
        - connect Alpha from sources (layeredTexture outAlpha for example)
        - test performance & memory on class vs. procedural functions
    '''
    #=========================================================================#
    # Primary Methods ========================================================#
    #=========================================================================#
    def __init__(self, profile = None, exclusions=None):
        '''
        Create gammaCorrect object, set basic attributes.
        Decompose profile into convenient attributes for clarity.
        '''
        self.verbose = 1 # 1:errors 2: warn 3: info 4: debug
        self.width = 60  # print width
        
        if not (exclusions): # use default
            exclusions = defaultExclusions()
        if not (profile):    # use default
            profile = defaultProfile()
        
        self.exclusions = exclusions
        self.profile = profile
        self.target = 'all'

        # decompose profile structure into sensible variables
        self.gNodeType = self.profile['gNodeType'][0]
        self.gAttrsIn  = self.profile['gAttrs'][0]
        self.gAttrsOut = self.profile['gAttrs'][1]
        self.gAttr     = self.profile['gSetting'][0]
        self.gLUT      = self.profile['gSetting'][1][0]
        self.gLUTi     = self.profile['gSetting'][1][1]
    def info(self, level=1, msg=None):
        '''
        1: errors
        2: warnings
        3: info
        4: debug
        '''
        status = None
        if   level == 5:
            status = 'DEBUG2'
        elif level == 4:
            status = 'DEBUG'
        elif level == 3:
            status = 'INFO'
        elif level == 2:
            status = 'WARN'
        elif level == 1:
            status = 'ERROR'
        else:
            pass
        if   self.verbose >= level:
            print '# {0:<6} ({1}): {2}'.format(status,self.__class__.__name__,msg)
    def preRenderCall(self):
        '''
        Queries target nodes and attributes, then calls correction for each.
        Restores initial selection after operation.
        '''
        initSel = mc.ls(sl=True)
        dstNodes = self.returnTargetMaterials()
        if dstNodes:
            for dstNode in dstNodes:
                dstAttrs = self.returnColorAttrs(dstNode)
                self.info(4,'{0:10}: {1:10}'.format('MAT',dstNode))
                self.info(4,'{0:10}: {1:10}'.format('ATTRS',dstAttrs))
                self.correctAttrs(dstNode,dstAttrs)
        if len(initSel):
            mc.select(initSel,replace=True,noExpand=True)
    def postRenderCall(self):
        '''
        Queries correction nodes in use and then removes them.
        Restores initial selection of nodes that still exist.
        '''
        initSel = mc.ls(sl=True)
        dstNodes = self.returnTargetMaterials()
        if dstNodes:
            gNodes = list()
            for dstNode in dstNodes:
                dstAttrs = self.returnColorAttrs(dstNode)
                for dstAttr in dstAttrs:
                    matPlugIn = mc.connectionInfo('{0}.{1}'.format(dstNode,dstAttr), sourceFromDestination=True)
                    inPlugNode = '{0}'.format(matPlugIn).partition('.')[0]
                    if mc.objExists(inPlugNode):
                        nType = mc.objectType(inPlugNode)
                        for gType in self.profile['gNodeType']:
                            if nType == gType:
                                gNodes.append(inPlugNode)
            self.removeNodes(gNodes)
        initSel[:] = [x for x in initSel if mc.objExists(x)]
        if len(initSel):
            mc.select(initSel,replace=True,noExpand=True)
    def insertTargetgNodes(self):
        '''
        Target always uses selection, even if none.
        Target nodes are presumed to be non-linear (unless excluded), therefore
        each target receives correction on its output.
        '''
        initSel = mc.ls(sl=True)
        shdrs = self.listShaders()
        shdrs = mc.ls(shdrs, sl=True)
        for shdr in shdrs:
            srcAttrs = self.returnColorAttrs(shdr, isInput=False, exclude=True, uac=False)
            for srcAttr in srcAttrs:
                srcPlug = '{0}.{1}'.format(shdr,srcAttr)
                srcPlug = self.excludePlug(srcPlug,con='src')
                if (srcPlug):
                    dstPlugs = mc.connectionInfo(srcPlug,dfs=True)
                    for dstPlug in dstPlugs:
                        dstPlug = self.excludePlug(dstPlug,con='dst')
                        if dstPlug:
                            dstNode = dstPlug.partition('.')[0]
                            dstAttr =  dstPlug.replace('{0}.'.format(dstNode),'')
                            self.info(3,'{0:-<{2}}->{1}'.format('insert: '+srcPlug, dstNode+'.'+dstAttr,self.width))
                            self.insertNode(dstNode, dstAttr, shdr, srcAttr)
        if len(initSel):
            mc.select(initSel,replace=True,noExpand=True)
    def removeTargetgNodes(self):
        '''
        target 'auto' affects selected nodes only (if any), else all if none.
        target 'all'  affects all nodes always (but still obeys exclusions).
        '''
        initSel = mc.ls(sl=True)
        gNodes = self.listgNodes()

        if not (self.target == 'auto' and not len(mc.ls(gNodes, sl=True))):
            gNodes = mc.ls(gNodes,sl=True)
        
        self.removeNodes(gNodes)
        initSel[:] = [x for x in initSel if mc.objExists(x)]
        if len(initSel):
            mc.select(initSel,replace=True,noExpand=True)
    def setTargetgNodes(self):
        '''
        Query target correction nodes and set them to current profile.
        ''' 
        gNodes = self.returnTargetgNodes()
        if (gNodes):
            self.setgNodes(gNodes)
    
    #=========================================================================#
    # Utility Methods ========================================================#
    #=========================================================================#
    def setLUT(self, gLUT=1.7,displayLUT=False,inputLUT=False):
        '''
        Sets display Look-Up Table/gamma and/or input LUT.
        For simplicity, display LUTs are inferred from a gamma value.
        g[~2.2] = sRGB
        g[~1.7] = rec709
        g[~1.0] = disabled / linear / no LUT
        '''
        gLUTi = (1.0/gLUT)
        # modify profile
        self.profile['gSetting'][1] = (gLUT,gLUTi)
        self.gLUT      = gLUT
        self.gLUTi     = gLUTi
        
        if (inputLUT):
            self.setTargetgNodes()
        if (displayLUT):
            if mc.allNodeTypes().count('viewColorManager'):
                # use floating point images
                mc.optionVar(iv=('renderViewImageFormat',1));
                
                tol = 0.1
                LUT = (2,3)
                if gLUT <= (2.2+tol) and gLUT >= (2.1-tol):
                    LUT = (2,3) # linear / sRGB
                elif gLUT <= (1.9+tol) and gLUT >= (1.6-tol):
                    LUT = (2,5) # linear / rec709
                elif gLUT <= (1.0+tol) and gLUT >= (1.0-tol):
                    LUT = (2,2) # linear / linear (no LUT)
                    
                vcms = mc.ls(type='viewColorManager')
                for vcm in vcms:
                    if mc.getAttr( '{0}.imageColorProfile'.format(vcm)) != LUT[0]:
                        mc.setAttr( '{0}.imageColorProfile'.format(vcm),   LUT[0])
                    if mc.getAttr( '{0}.displayColorProfile'.format(vcm)) != LUT[1]:
                        mc.setAttr( '{0}.displayColorProfile'.format(vcm), LUT[1])
            else:
                self.info(1, 'viewColorManager not supported. Use alternate.')
    def setgNodes(self, gNodes):
        '''
        Set profile on list of correction nodes to the current LUT
        '''
        if gNodes:
            for gNode in gNodes:
                if mc.objExists(gNode):
                    # set LUT on gNode
                    for gAttr in self.gAttr:
                        mc.setAttr('{0}.{1}'.format(gNode,gAttr),self.gLUTi)
    def listgNodes(self):
        '''
        List all correction nodes except those excluded by name.
        (exclusions by type would be an ironic waste)
        TODO: test more thoroughly
        '''
        gNodes = mc.ls(type=[self.gNodeType])
        gNodes = self.excludeNodes(gNodes,'name')
        return gNodes
    def returnTargetgNodes(self):
        '''
        list target correction nodes.
        target 'auto' affects selected shaders only (if any), else all if none.
        target 'all'  affects all shaders always (but still obeys exclusions).
        '''
        gNodes = self.listgNodes()
        
        if self.target == 'all':
            return gNodes
        else:
            if len(mc.ls(gNodes, sl=True)):
                return mc.ls(gNodes, sl=True)
            elif self.target == 'auto':
                return gNodes
    
    #=========================================================================#
    # Exclusion Methods ======================================================#
    #=========================================================================#
    def excludeNodes(self, nodes, mode = 'type', con='srcdst'):
        '''
        input  : node type/name
        output : node type/name
        most simple exclusion.
        Takes list of nodes names/types and returns a list of allowed nodes by
        node name and type respectively.
        '''
        # if key is in the exclusions dictionary
        if mode in self.exclusions:
            exclList = list()
            for node in nodes:
                exclItem = [x for (x,y,z) in self.exclusions[mode] \
                            if (x=='{0}'.format(node) and y == '*') \
                            and con.count(z)]
                if (exclItem):
                    exclList.extend(exclItem)
            self.info(4,'excluded node {0:>4}s: {1}'.format(mode,exclList))
            # check node exclusions
            nodes[:] = [node for node in nodes if not exclList.count(node)]
        return nodes
    def excludeAttrs(self, node, attrs, mode = 'type', con='srcdst'):
        '''
        input  : node type/name + attrs
        output : attrs
        For a given node name/type and attributes, return those attributes
        which are allowed. attribute names and types are synonymous.
        '''
        # if key is in the exclusions dictionary
        if mode in self.exclusions:
            exclAttrs = list()
            for at in attrs:
                exclItem = [y for (x,y,z) in self.exclusions[mode] \
                            if ((y=='{0}'.format(at) and \
                                (x == '*' or x == '{0}'.format(node)))\
                                or (x=='{0}'.format(node) and y =='*')) \
                            and con.count(z)]
                if (exclItem):
                    exclAttrs.extend(exclItem)
            self.info(4,'excluded attr {0:>4}s: {1}'.format(mode,exclAttrs))
            # check attribute exclusions
            attrs[:] = [at for at in attrs \
                        if not exclAttrs.count(at) and not exclAttrs.count('*')]
        return attrs
    def excludePlug(self,plug, con='srcdst'):
        '''
        input  : plug
        output : plug
        
        Most common exclusions are shader inputs (where a material type/name is
        specified and/or attributes are excluded). Generic node types/names are
        allowed for target correction/removal.
        
        Note: Material types are always skipped as source, (output is always
            linear, even if plugged into other shaders). If shader inputs are
            nonlinear, compensating with shader output just clouds the problem,
            and is therefore not allowed.
        TODO: If there's one section of code that could use thorough bug
            testing: it's here. Exclusions are cryptic, even without regex.
        '''
        matTypes = mc.listNodeTypes('shader')
        matNodes = mc.ls(type=matTypes)
        node = plug.partition('.')[0]
        attr = plug.replace('{0}.'.format(node),'')
        #attr = plug.rpartition('.')[2] # last attribute (in case of multi)
        
        # node exclusions
        if (node and attr):
            if 'type' in self.exclusions:
                exclTypes = [x for (x,y,z) in self.exclusions['type'] \
                            if (((x=='{0}'.format(mc.nodeType(node)) and y == '*' ) \
                            or ((x=='{0}'.format(mc.nodeType(node)) and y == attr) \
                            and y==attr)) and con.count(z))]
                node = node if (node) and \
                (not exclTypes.count('{0}'.format(mc.nodeType(node)))) else None
        if (node and attr):
            if 'name' in self.exclusions:
                exclNames = [x for (x,y,z) in self.exclusions['name'] \
                            if ((x==node and y == '*') \
                            or (x==node and y==attr)) \
                            and con.count(z)]
                node = node if (node) and \
                (not exclNames.count(node)) else None
        
        # exclusions of shader-shader connections
        if (node and attr):
            node = node if (node) and \
            (not (matTypes.count(mc.nodeType(node)) and con.count('src')) ) \
            else None
        
        
        # attr exclusions
        if (node and attr):
            if 'type' in self.exclusions:
                exclTypes = [y for (x,y,z) in self.exclusions['type'] \
                    if (y==attr and (x=='*' or x=='{0}'.format(mc.nodeType(node)))) and con.count(z)]
                attr = attr if (attr) \
                and (not exclTypes.count(attr)) and (not exclTypes.count('*')) else None
        if (node and attr):
            if 'name' in self.exclusions:
                exclNames = [y for (x,y,z) in self.exclusions['name'] \
                    if (y==attr and (x=='*' or x==node)) and con.count(z)]
                attr = attr if (attr) \
                and (not exclNames.count(attr)) and (not exclNames.count('*')) else None
        
        if not (node and attr):
            self.info(4,'excluded plug: {0}'.format(plug))
        else:
            return plug
    
    #=========================================================================#
    # General Purpose Methods ================================================#
    #=========================================================================#
    def returnTargetMaterials(self):
        '''
        Return materials from input list (usually selection), if list is empty,
        query all materials instead.
        Materials are shaders applied to shadingGroups (shaders which are not
        utility nodes).
        '''
        materials = self.listMaterials()
        if self.target == 'all':
            return materials
        elif self.target == 'auto':
            if not len(mc.ls(materials, sl=True)):
                return materials
            else:
                return mc.ls(materials, sl=True)
        else: # force selection
            return mc.ls(materials, sl=True)
    def listMaterials(self):
        '''
        Return non-excluded "Materials"
        '''
        mats = list()
        types = mc.listNodeTypes('shader')
        types = self.excludeNodes(types,'type')
        mats.extend(mc.ls(type=types))
        # remove duplicates thanks to Autodesk bug [ls -type "lambert"]
        mats = list(set(mats))
        if len(mats):
            mats = self.excludeNodes(mats,'name')
        return mats
    def listShaders(self):
        '''
        Return most nodes which can be used in a shading network.
        '''
        shdrs = list()
        types = list()
        types.extend(mc.listNodeTypes('rendernode'))
        types.extend(mc.listNodeTypes('utility'))
        types.extend(mc.listNodeTypes('texture'))
        types.extend(mc.listNodeTypes('light'))
        types.extend(mc.listNodeTypes('shader'))
        types = self.excludeNodes(types,'type')
        
        shdrs.extend(mc.ls(type=types))
        shdrs = self.excludeNodes(shdrs,'name')
        return shdrs
    def returnColorAttrs(self, node, isInput=True ,exclude=True,uac=True):
        '''
        Return list of colors which need correction
        uac = used as color
        '''
        # list all attrs
        attrs = mc.listAttr(node,
                    connectable = True,
                    unlocked    = True,
                    write       = isInput,
                    read    = not isInput,
                    settable    = isInput,
                    visible     = True,
                    multi       = True)
        # isolate color attrs
        attrs = [at for at in attrs \
                    if mc.getAttr('{0}.{1}'.format(node,at),type=True) == 'float3']
        # note: multi-attributes fail, use rpartition to find the end color only
        if uac:
            attrs = [at for at in attrs \
                if mc.attributeQuery('{0}'.format(at).rpartition('.')[2],node='{0}'.format(node),usedAsColor=True)]

        if exclude:
            attrs = self.excludeAttrs(mc.nodeType(node), attrs, 'type')
            attrs = self.excludeAttrs(node, attrs, 'name')
        return attrs
    def correctAttrs(self,dstNode, dstAttrs):
        for attr in dstAttrs:
            plugMat = '%s.%s' % (dstNode,attr)
            plugCxn = mc.connectionInfo('%s'% (plugMat) ,sfd=True)
            
            if len(plugCxn):
                self.correctInput(dstNode,attr)
            else:
                self.correctSolid(dstNode,attr)
    def correctInput(self, dstNode, dstAttr):
        '''
        TODO:
            - check with animated colors
        '''
        dstPlug  = '{0}.{1}'.format(dstNode,dstAttr)
        srcPlug = mc.connectionInfo(dstPlug ,sfd=True)
        # test exclusions WITH destination type
        srcPlug = self.excludePlug(srcPlug)
        self.info(4,'plug: {0}'.format(srcPlug))
        if (srcPlug):
            srcNode = srcPlug.partition('.')[0]
            srcAttr = srcPlug.replace('{0}.'.format(srcNode),'')
            self.insertNode(dstNode, dstAttr,srcNode,srcAttr)
    def correctSolid(self,dstNode, dstAttr):
        '''
        
        TODO:
            - thoroughly bug test when modifying profile structure
            - could possibly share some code with insertNode
        '''
        dstPlug  = '{0}.{1}'.format(dstNode,dstAttr)
        
        # tolerance
        tol = 0.0001
        rgb = mc.getAttr('{0}.{1}'.format(dstNode,dstAttr))[0]
        if  ((rgb[0] <= 0.0+tol) or (rgb[0] >= 1.0-tol)) and \
            ((rgb[1] <= 0.0+tol) or (rgb[1] >= 1.0-tol)) and \
            ((rgb[2] <= 0.0+tol) or (rgb[2] >= 1.0-tol)):
            self.info(4,'{0:-<{4}}->({1:<0.3f},{2:<0.3f},{3:<0.3f})'.format(dstPlug +' (skipping)',rgb[0],rgb[1],rgb[2],self.width))
        else:
            self.info(4,'{0:-<{4}}->({1:<0.3f},{2:<0.3f},{3:<0.3f})'.format(dstPlug +' (correcting)',rgb[0],rgb[1],rgb[2],self.width))
            gNode = mc.shadingNode(self.gNodeType,asUtility=True,name='{0}__{1}GC'.format(dstNode,dstAttr))
            
            # set LUT on gNode
            for gAttr in self.gAttr:
                mc.setAttr('{0}.{1}'.format(gNode,gAttr),self.gLUTi)
            
            # set rgba on gNode
            for gAttrIn in self.gAttrsIn:
                gPlugIn = '{0}.{1}'.format(gNode,gAttrIn)
                if mc.getAttr(gPlugIn,type=True) == 'float3':
                    self.info(4,'{0:-<{4}}->({1:<0.3f},{2:<0.3f},{3:<0.3f})'.format(gPlugIn,rgb[0], rgb[1], rgb[2],self.width))
                    mc.setAttr(gPlugIn,rgb[0], rgb[1], rgb[2],type='double3')
                elif mc.getAttr(gPlugIn,type=True) == 'float':
                    self.info(4,'{0:-<{2}}->({1:<0.3f})'.format(gPlugIn,1.0,self.width))
                    mc.setAttr(gPlugIn,1.0)
                else:
                    pass
            
            # connect gNode to dstNode
            for gAttrOut in self.gAttrsOut:
                gPlugOut = '{0}.{1}'.format(gNode,gAttrOut)
                if mc.getAttr(gPlugOut,type=True) == 'float3' \
                and mc.getAttr(dstPlug ,type=True) == 'float3':
                    self.info(3,'{0:-<{2}}->{1}'.format(gPlugOut,dstPlug ,self.width))
                    mc.connectAttr(gPlugOut,
                                   dstPlug ,f=True)
                elif mc.objExists('{0}A'.format(dstPlug )) \
                and mc.getAttr(gPlugOut,type=True) == 'float' \
                and mc.getAttr('{0}A'.format(dstPlug ),type=True) == 'float':
                    self.info(4,'{0:-<{2}}->{1}'.format(gPlugOut,dstPlug ,self.width))
                    mc.connectAttr(gPlugOut,
                                   '{0}A'.format(dstPlug ),f=True)
                else:
                    pass
    def insertNode(self, dstNode, dstAttr,srcNode,srcAttr):
        '''
        dstNode = (_._)->gNode->(*._)
        dstAttr = (_._)->gNode->(_.*)
        srcNode = (*._)->gNode->(_._)
        srcAttr = (_.*)->gNode->(_._)
        
        TODO:
            - thoroughly bug test when modifying profile structure
            - could possibly share some code with correctSolid
            - could/should be made more generic, for repurposing later.
        '''
        dstPlug   = '{0}.{1}'.format(dstNode ,dstAttr)
        srcPlug = '{0}.{1}'.format(srcNode,srcAttr)
        self.info(3,'{0:-<60}->{1}'.format(srcPlug,dstPlug ))
        
        ### prevention check 1 : linear file textures ###
        if mc.attributeQuery('fileTextureName',exists=True,node=srcNode):
            fn = mc.getAttr('{0}.fileTextureName'.format(srcNode))
            if len(fn):
                ext = fn.rpartition('.')[2]
                self.info(3,'{0} file extension is:({1})'.format(srcNode,ext))
                if ext == 'hdr' or ext == 'exr':
                    return None
        ### prevention check 2: recursion ###
        for gNodeType in self.profile['gNodeType']:
            if mc.nodeType(srcNode) == gNodeType \
            or mc.objectType(dstNode ) == gNodeType:
                return None
        
        ### prevention check 3: duplicates/recycling ###
        gNode = ''
        srcPlugDstPlugs = mc.connectionInfo(srcPlug,dfs=True)
        for srcdstPlug in srcPlugDstPlugs:
            for gNodeType in self.profile['gNodeType']:
                if mc.objectType(srcdstPlug) == gNodeType:
                    gNode = srcdstPlug.partition('.')[0]
        
        # create correction node
        if not mc.objExists(gNode):
            gNode = mc.shadingNode(self.gNodeType,asUtility=True,
                                   name='{0}GC'.format(srcNode))
            # set LUT on gNode
            for gAttr in self.gAttr:
                mc.setAttr('{0}.{1}'.format(gNode,gAttr),self.gLUTi)
        # make connections
        # connect src to gNode
        for gAttrIn in self.gAttrsIn:
            gPlugIn = '{0}.{1}'.format(gNode,gAttrIn)
            if mc.getAttr(gPlugIn,type=True) == 'float3' \
            and mc.getAttr(srcPlug,type=True) == 'float3':
                if not mc.isConnected(srcPlug,gPlugIn):
                    self.info(3,'{0:-<{2}}->{1}'.format(srcPlug,gPlugIn,self.width))
                    mc.connectAttr(srcPlug,gPlugIn)
            elif mc.getAttr(gPlugIn,type=True) == 'float' \
            and mc.getAttr(srcPlug,type=True) == 'float' \
            and mc.objExists('{0}A'.format(mSrcPlug)):
                if not mc.isConnected('{0}A'.format(srcPlug),gPlugIn):
                    self.info(4,'{0:-<{2}}->{1}'.format(srcPlug,gPlugIn,self.width))
                    mc.connectAttr('{0}A'.format(srcPlug),gPlugIn)
            else:
                pass
        # connect gNode to dstNode 
        for gAttrOut in self.gAttrsOut:
            gPlugOut = '{0}.{1}'.format(gNode,gAttrOut)
            if mc.getAttr(gPlugOut,type=True) == 'float3' \
            and mc.getAttr(dstPlug ,type=True) == 'float3':
                self.info(3,'{0:-<{2}}->{1}'.format(gPlugOut,dstPlug ,self.width))
                mc.connectAttr(gPlugOut,dstPlug ,f=True)
            elif mc.objExists('{0}A'.format(dstPlug )) \
            and mc.getAttr('{0}A'.format(dstPlug ),type=True) == 'float' \
            and mc.getAttr(gPlugOut,type=True) == 'float':
                self.info(4,'{0:-<{2}}->{1}'.format(gPlugOut,dstPlug ,self.width))
                mc.connectAttr(gPlugOut,'{0}A'.format(dstPlug ),f=True)
            else:
                pass
    def removeNodes(self, gNodes):
        '''
        TODO:
            - Possibly exclude Nodes from removal, optionally
            - Possibly skip nodes with non-profile attributes
        '''
        gNodes = list(set(gNodes))
        self.info(4,'removing nodes: {0}'.format(gNodes))
        for gNode in gNodes:
            nType = mc.nodeType(gNode)
            
            for gNodeType in self.profile['gNodeType']:
                if nType == gNodeType:
                    gDstPlugs = list()
                    gSrcPlugs = list()
                    for gAttrOut in self.gAttrsOut:
                        gDstPlugList = mc.connectionInfo('{0}.{1}'.format(gNode,gAttrOut),dfs=True)
                        for gDstPlug in gDstPlugList:
                            if len(gDstPlug):
                                gDstPlugs.append(gDstPlug)
                    for gAttrIn in self.gAttrsIn:
                        gSrcPlug = mc.connectionInfo('{0}.{1}'.format(gNode,gAttrIn),sfd=True)
                        if len(gSrcPlug):
                            gSrcPlugs.append(gSrcPlug)
                    
                    self.info(4,'{0:<10}: {1}'.format('gDstPlugs',gDstPlugs))
                    self.info(4,'{0:<10}: {1}'.format('gSrcPlugs',gSrcPlugs))
                    
                    # has color input(s)
                    if len(gSrcPlugs):
                        # each destination
                        for gDstPlug in gDstPlugs:
                            # each input connection
                            for gSrcPlug in gSrcPlugs:
                                if mc.getAttr(gSrcPlug,type=True) == \
                                              mc.getAttr(gDstPlug,type=True):
                                    if not mc.isConnected(gSrcPlug,gDstPlug):
                                        self.info(4,'{0:-<{2}}->{1}'.format(gSrcPlug,gDstPlug,self.width))
                                        mc.connectAttr(gSrcPlug,gDstPlug,force=True)
                                    
                    # is solid color
                    else:
                        # each destination
                        for gDstPlug in gDstPlugs:
                            # each input color
                            for gAttrIn in self.gAttrsIn:
                                for gAttrOut in self.gAttrsOut:
                                    gOutPlug = '{0}.{1}'.format(gNode,gAttrOut)
                                    
                                    self.info(4,'{0:-<{2}}->{1}'.format(gOutPlug,gDstPlug,self.width))
                                    if mc.isConnected(gOutPlug,gDstPlug):
                                        mc.disconnectAttr(gOutPlug,gDstPlug)
                                        
                                        if mc.getAttr(gOutPlug,type=True) == 'float3'\
                                        and mc.getAttr(gDstPlug,type=True) == 'float3':
                                            rgb = mc.getAttr('{0}.{1}'.format(gNode,gAttrIn))[0]
                                            self.info(3,'{0:-<{4}}->({1:<0.3f},{2:<0.3f},{3:<0.3f})'.format(gDstPlug,rgb[0], rgb[1], rgb[2],self.width))
                                            mc.setAttr(gDstPlug,rgb[0], rgb[1], rgb[2],type='double3')
                                        elif mc.getAttr(gOutPlug,type=True) == 'float'\
                                        and mc.getAttr(gDstPlug,type=True) == 'float'\
                                        and mc.objExists('{0}.{1}A'.format(gNode,gAttrIn)):
                                            a = mc.getAttr('{0}.{1}A'.format(gNode,gAttrIn))
                                            self.info(4,'{0:-<{2}}->({1:<0.3})'.format(gDstPlug,a,self.width))
                                            mc.setAttr(gDstPlug,a)
                                        else:
                                            self.info(1,'wrong types: {0} and {1}'.format(mc.getAttr(gSrcPlug,type=True),mc.getAttr(gDstPlug,type=True)))
                    mc.delete(gNode)
        return None
    
    #=========================================================================#

class UI(object):
    '''
    user interface class. use wLinearWorkflow.loadUI() for simple creation
    TODO:
        - optionVar preferences.
        - load/save exclusions
    '''
    def __init__(self,windowName='wLinearWorkflow'):
        '''
        Constructor
        '''
        self.verbose = 5
        self.winName = windowName
        self.iconPath = '{0}/icons/'.format(__path__[0])
        # pickle path
        self.pp = '{0}/prefs/{1}.pkl'.format(__path__[0],os.getlogin())
        if not os.path.isdir('{0}/prefs'.format(__path__[0])):
            os.mkdir('{0}/prefs'.format(__path__[0]))
        self.pkl = open(self.pp,'wb')
        
        gc = gammaCorrect()
        gc.target = 'auto'
        
        if mc.window(self.winName, exists=True):
            mc.deleteUI(self.winName)
            #mc.windowPref(self.winName,remove=True)
        #======================================================================
        #  Create the GUI
        #======================================================================
        mc.window(self.winName, menuBar=True, title='{0}'.format(self.winName),
                  width=140, height=100,sizeable=False)
        mc.menu(helpMenu=True,label='help')
        mc.menuItem(label='print help info',
                    c=(self.buttonCB_help))
        bwidth = 50
        bpad = 5
        
        FrL01   = mc.frameLayout(collapsable=False,
                                 label='',labelVisible=True)
        ClLO2a  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        #TL00a   = mc.tabLayout()
        
        FrLO3I  = mc.frameLayout(borderStyle='out',labelVisible=True,label='Automatic',labelIndent=10)
        ClLO4I  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        mc.text(label='Pre/Post-Render',font='boldLabelFont')
        RCLO5I  = mc.rowColumnLayout(numberOfColumns=2,
                                     columnAttach=[(1,'left',bpad),(2,'right',bpad)])
        mc.iconTextButton(label='On',image='editRenderPass.png',
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(setup,True))
        mc.iconTextButton(label='Off',image='deleteRenderPass.png',
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(setup,False))
        mc.setParent('..')
        mc.setParent('..')
        mc.setParent('..') # end auto setup
        
        
        FrLO3T  = mc.frameLayout(collapsable=True,collapse=False,label='Tools')
        ClLO4T  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        FrLO5Tb  = mc.frameLayout(borderStyle='out',labelVisible=False)
        ClLO6Tb  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        mc.text(label='Network',font='boldLabelFont')
        RCLO7Tb  = mc.rowColumnLayout(numberOfColumns=2,
                                      columnAttach=[(1,'left',bpad),(2,'right',bpad)])
        mc.iconTextButton(label='On',image='{0}wGamma_shaders_on.xpm'.format(self.iconPath),
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(self.buttonCB_autocorrect))
        mc.iconTextButton(label='Off',image='{0}wGamma_shaders_off.xpm'.format(self.iconPath),
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(self.buttonCB_autocleanup))
        mc.setParent('..')
        mc.setParent('..')
        mc.setParent('..')
        
        FrLO5Tc  = mc.frameLayout(borderStyle='out',labelVisible=False)
        ClLO6Tc  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        mc.text(label='Target',font='boldLabelFont')
        RCLO7Tc  = mc.rowColumnLayout(numberOfColumns=2,
                                      columnAttach=[(1,'left',bpad),(2,'right',bpad)])
        mc.iconTextButton(label='insert',image='{0}wGamma_insert.xpm'.format(self.iconPath),
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(self.buttonCB_targetinsert))
        mc.iconTextButton(label='remove',image='{0}wGamma_remove.xpm'.format(self.iconPath),
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(self.buttonCB_targetremove))
        mc.setParent('..')
        mc.setParent('..')
        mc.setParent('..')
        
        mc.setParent('..')
        mc.setParent('..') # end tools
        
        #mc.tabLayout(TL00a, edit=True, tabLabel=[(FrLO3I,'Auto'),(FrLO3T,'Tools')])
        #mc.setParent('..')
        
        FrLO3O  = mc.frameLayout(collapsable=True,collapse=False,label='Options')
        ClLO4O  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        FrLO5O  = mc.frameLayout(collapsable=True,collapse=False,label='Profile')
        ClLO6O  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        FrLO5Ta  = mc.frameLayout(borderStyle='out',labelVisible=False)
        ClLO6Ta  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        mc.text(label='Display LUT',font='boldLabelFont')
        RCLO7Ta  = mc.rowColumnLayout(numberOfColumns=2,
                                      columnAttach=[(1,'left',bpad),(2,'right',bpad)])
        mc.popupMenu()
        mc.menuItem(label='select display LUT node',
                    c=partial(self.buttonCB_selDVCM))
        mc.iconTextButton(label='Enable',image='{0}wGamma_display_on.xpm'.format(self.iconPath),
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(self.buttonCB_LUTon))
        mc.iconTextButton(label='Disable',image='{0}wGamma_display_off.xpm'.format(self.iconPath),
                          style='iconAndTextVertical',width=bwidth,
                          c=partial(self.buttonCB_LUToff))
        mc.setParent('..')
        mc.setParent('..')
        mc.setParent('..')
        
        #mc.text(label='(drop down list)?')
        #mc.text(label='(set gamma)?')
        FrLO5Tb  = mc.frameLayout(borderStyle='out',labelVisible=False)
        ClLO6Tb  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        mc.text(label='Input LUT',font='boldLabelFont')
        self.gSlider = mc.floatSliderGrp( label='g', field=True,
                           adjustableColumn=3,
                           columnWidth=[(1,10),(2,30),(3,50)],
                           minValue=1.0, maxValue=3.0,
                           fieldMinValue=0.001, fieldMaxValue=10.0,
                           sliderStep=0.1, value=2.2,
                           changeCommand=partial(self.sliderCB_gamma))
        mc.popupMenu()
        mc.menuItem(label='select input LUT nodes',
                    c=partial(self.buttonCB_selgNodes))
        mc.setParent('..')
        mc.setParent('..')
        self.linkLUT = mc.checkBox(label='link I/O LUTs', value=True,
            ann='g[1.0]=noLUT, g[1.7]=rec709, g[2.2]=sRGB')
        mc.setParent('..')
        mc.setParent('..') # end Profile
        
        FrLO5O  = mc.frameLayout(collapsable=True,collapse=False,label='Target')
        ClLO6O  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        self.targetMode = mc.radioButtonGrp(vertical=True,
                          numberOfRadioButtons=3,columnAlign=[1,'center'],
                          changeCommand=partial(self.buttonCB_targetMode),
                          select=1,labelArray3=['Automatic','Selection Only','Always All'])
        mc.setParent('..')
        mc.setParent('..') # end Target
        
        FrLO5O  = mc.frameLayout(collapsable=True,collapse=False,borderStyle='out',label='Exclusions')
        ClLO6O  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        mc.button(label='Exclusions UI',c=partial(self.exclusionsUI,'Exclusions'))
        
        mc.setParent('..')
        mc.setParent('..') # end exclusions
        mc.setParent('..')
        mc.setParent('..') # end options
        #mc.setParent(FrL01)
        
        self.psave(gc)
        mc.showWindow(self.winName)
    def exclusionsUI(self,windowName,*args):
        self.exclWin = windowName
        if mc.window(self.exclWin, exists=True):
            mc.deleteUI(self.exclWin)
            #mc.windowPref(self.exclWin,remove=True)
        
        # pickle load
        gc = self.pload()

        mc.window(self.exclWin, menuBar=True, title='{0}'.format(self.exclWin),
                  width=400, height=300,sizeable=False)
        FrL01   = mc.frameLayout(collapsable=False,
                                 label='',labelVisible=False)
        ClLO2a  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        mc.rowLayout(numberOfColumns=4,adjustableColumn=2)
        mc.text(label='preset')
        self.excl_presets = mc.optionMenu()
        # load presets TODO: actually load presets
        mc.menuItem(label='default')
        mc.menuItem(label='preset1')
        mc.iconTextButton(image='fileOpen.png')
        mc.iconTextButton(image='fileSave.png')
        
        mc.setParent('..')
        
        FrLO7Oa  = mc.frameLayout(labelVisible=True,borderStyle='out',label='Exclude by Node Type:') # Type
        ClLO8Oa  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        mc.text(font='smallFixedWidthFont', align='left',
                label='{0:>4}{2:^54}{1:<7}'.format('dst','src','{0:>20}.{1:<20}'.format('(nodeType)','(attribute)')))
        self.exclTSL_T = mc.textScrollList(height=100,allowMultiSelection=True,
            font='smallFixedWidthFont')
        
        RCLO7OaT = mc.formLayout()
        b08aT1   = mc.iconTextButton(image='setEdAddCmd.png',height=32,width=32,
                          style='iconOnly',
                          c=partial(self.buttonCB_exclTSL_add,'type'))
        b08aT2   = mc.iconTextButton(image='setEdEditMode.png',height=32,width=32,
                          style='iconOnly',
                          c=partial(self.buttonCB_exclTSL_edit,'type'))
        b08aT3   = mc.iconTextButton(image='setEdRemoveCmd.png',height=32,width=32,
                          style='iconOnly',
                          c=partial(self.buttonCB_exclTSL_remove,'type'))
        mc.formLayout(RCLO7OaT,edit=True,
                      attachForm=[(b08aT1,'left',20),(b08aT3,'right',20),
                                  (b08aT1,'top',0),(b08aT2,'top',0),(b08aT3,'top',0),
                                  (b08aT1,'bottom',0),(b08aT2,'bottom',0),(b08aT3,'bottom',0)],
                      attachControl=[(b08aT2,'left',20,b08aT1),(b08aT2,'right',20,b08aT3)])
        mc.setParent('..')
        mc.setParent('..')
        mc.setParent('..')# end exclusion by type
        
        mc.separator(horizontal=True)
        
        FrLO7Ob  = mc.frameLayout(labelVisible=True,borderStyle='out',label='Exclude by Node Name:') # Name
        ClLO8Ob  = mc.columnLayout(adjustableColumn=True,columnAttach=('both',0))
        
        mc.text(font='smallFixedWidthFont', align='left',
                label='{0:>4}{2:^54}{1:<7}'.format('dst','src','{0:>20}.{1:<20}'.format('(nodeName)','(attribute)')))
        self.exclTSL_N = mc.textScrollList(height=100,allowMultiSelection=True,
            font='smallFixedWidthFont')
        
        RCLO7OaN = mc.formLayout()
        b08aN1   = mc.iconTextButton(image='setEdAddCmd.png',height=32,width=32,
                          style='iconOnly',
                          c=partial(self.buttonCB_exclTSL_add,'name'))
        b08aN2   = mc.iconTextButton(image='setEdEditMode.png',height=32,width=32,
                          style='iconOnly',
                          c=partial(self.buttonCB_exclTSL_edit,'name'))
        b08aN3   = mc.iconTextButton(image='setEdRemoveCmd.png',height=32,width=32,
                          style='iconOnly',
                          c=partial(self.buttonCB_exclTSL_remove,'name'))
        mc.formLayout(RCLO7OaN,edit=True,
                      attachForm=[(b08aN1,'left',20),(b08aN3,'right',20),
                                  (b08aN1,'top',0),(b08aN2,'top',0),(b08aN3,'top',0),
                                  (b08aN1,'bottom',0),(b08aN2,'bottom',0),(b08aN3,'bottom',0)],
                      attachControl=[(b08aN2,'left',20,b08aN1),(b08aN2,'right',20,b08aN3)])
        mc.setParent('..')
        mc.setParent('..')
        mc.setParent('..') # end exclusion by name
        
        mc.setParent('..')
        mc.setParent('..') # end main
        
        self.psave(gc)
        self.exclTSL_update()
        
        mc.showWindow(self.exclWin)
    def info(self, level=1, msg=None):
        '''
        '''
        status = None
        if   level == 5:
            status = 'DEBUG2'
        elif level == 4:
            status = 'DEBUG'
        elif level == 3:
            status = 'INFO'
        elif level == 2:
            status = 'WARN'
        elif level == 1:
            status = 'ERROR'
        else:
            pass
        if   self.verbose >= level:
            print '# {0:<6} ({1}): {2}'.format(status,self.__class__.__name__,msg)
    
    #=========================================================================#
    # Interface Functions ====================================================#
    #=========================================================================#
    def pload(self):
        self.pkl = open(self.pp,'rb')
        gc = pickle.load(self.pkl)
        self.pkl.close()
        return gc
    def psave(self,gc):
        self.pkl = open(self.pp,'wb')
        pickle.dump(gc,self.pkl)
        self.pkl.close()
        del gc
    def buttonCB_help(self,*args):
        #help(self.__class__)
        help(self.__module__)
        mm.eval('print("// wLinearWorkflow help: see script editor //\\n")')
    def buttonCB_install(self):
        setup(True)
    def buttonCB_uninstall(self):
        setup(False)
    def buttonCB_autocorrect(self):
        gc = self.pload()
        gc.preRenderCall()
        del gc
    def buttonCB_autocleanup(self):
        gc = self.pload()
        gc.postRenderCall()
        del gc
    def buttonCB_targetinsert(self):
        gc = self.pload()
        gc.insertTargetgNodes()
        del gc
    def buttonCB_targetremove(self):
        gc = self.pload()
        gc.removeTargetgNodes()
        del gc
    def buttonCB_LUTon(self):
        isLinked = mc.checkBox(self.linkLUT,q=True,value=True)
        gLUT  = mc.floatSliderGrp(self.gSlider,q=True,value=True)
        gc = self.pload()
        gc.setLUT(gLUT,True,isLinked)
        self.psave(gc)
    def buttonCB_LUToff(self):
        isLinked = mc.checkBox(self.linkLUT,q=True,value=True)
        gc = self.pload()
        gc.setLUT(1.0,True,False)
        self.psave(gc)
    def buttonCB_selDVCM(self,*args):
        mc.select(mc.ls('defaultViewColorManager'))
    def buttonCB_selgNodes(self,*args):
        gc = self.pload()
        gNodes = gc.returnTargetgNodes()
        if (gNodes):
            mc.select(gNodes)
        self.psave(gc)
    def buttonCB_targetMode(self,*args):
        gc = self.pload()
        if mc.radioButtonGrp(self.targetMode, q=True,select=1) == 1:
            gc.target = 'auto'
        elif mc.radioButtonGrp(self.targetMode, q=True,select=3) == 3:
            gc.target = 'all'
        else: # mc.radioButtonGrp(self.targetMode, q=True,select=2):
            gc.target = 'force_selection'
        self.psave(gc)
    def sliderCB_gamma(self, *args):
        # load UI setting
        isLinked = mc.checkBox(self.linkLUT,q=True,value=True)
        gLUT  = mc.floatSliderGrp(self.gSlider,q=True,value=True)
        # load gc object
        gc = self.pload()
        # modify profile
        gc.setLUT(gLUT,isLinked,True)
        # save gc object
        self.psave(gc)
    def buttonCB_exclTSL_remove(self,mode,*args):
        TSL = self.exclTSL_T if mode == 'type' else self.exclTSL_N
        sel = mc.textScrollList(TSL, q=True, selectItem=True)
        if (sel):
            # remove from data
            rList = list()
            for item in sel:
                rList.append( self.exclTSL_item_info(item,mode) )
            self.exclTSL_del(rList)
            # remove from list
            for item in sel:
                #print('['+item+']')
                self.info(4,'removing exclusion: {0}'.format(self.exclTSL_item_info(item,mode)))
                mc.textScrollList(TSL, e=True, removeItem=item)
            self.exclTSL_update()
    def buttonCB_exclTSL_edit(self,mode,*args):
        gc = self.pload()
        TSL = self.exclTSL_T if mode == 'type' else self.exclTSL_N  
        sel = mc.textScrollList(TSL, q=True, selectItem=True)
        # actions!
        if (sel):
            for item in sel:
                print(self.exclTSL_item_info(item,mode))
        #self.exclTSL_edit_win()
        self.psave(gc)
        #self.exclTSL_update()
    def buttonCB_exclTSL_add(self,mode,*args):
        TSL = self.exclTSL_T if mode == 'type' else self.exclTSL_N
        self.exclTSL_add_win(mode)
    def buttonCB_exclAdd(self,mode,tN,tA,c1,c2,*args):
        '''
        '''
        node = mc.textField(tN,q=True,text=True)
        attr = mc.textField(tA,q=True,text=True)
        target =  'src' if mc.checkBox(c2,q=True,value=True) else ''
        target += 'dst' if mc.checkBox(c1,q=True,value=True) else ''
        gc = self.pload()
        
        if (node and attr and target):
            if mode in gc.exclusions:
                if not (gc.exclusions[mode].count((node,attr,target))):
                    gc.exclusions[mode].append((node,attr,target))
                else:
                    self.info(3,'skipping exclusion: ({0}.{1}) as {2}'.format(node,attr,target))
            else:
                gc.exclusions[mode]  = [(node,attr,target)]
        self.psave(gc)
        self.exclTSL_update()
    def exclTSL_item_info(self,item,mode):
        p1 = item.partition('(')
        p2 = p1[2].partition(')')
        
        target =  'src' if p2[2].strip().count('->') else ''
        target += 'dst' if p1[0].strip().count('->') else ''
        node = p2[0].partition('.')[0].strip()
        attr = p2[0].replace('{0}.'.format(node),'').strip()
        return [mode,node,attr,target]
    def exclTSL_add_win(self,mode='type'):
        w1 = 'addWindow'
        if mc.window(w1, exists=True):
            mc.deleteUI(w1)
            
        w1 = mc.window(w1,title='add Exclusion',height=60,width=400)
        f1 = mc.formLayout(numberOfDivisions=8)
        
        c1 = mc.checkBox( label='',ann='-->(*.*) destination' )
        tN = mc.textField(w= 50,text='NODE')
        t1 = mc.text(label='.')
        tA = mc.textField(w= 50,text='ATTR')
        c2 = mc.checkBox( label='',ann='(*.*)--> source' )
        
        b1 = mc.button(label='add',
                       c=partial(self.buttonCB_exclAdd,mode,tN,tA,c1,c2))
        
        mc.formLayout(f1,e=True,
            attachForm=[(c1,'top',10),(c2,'top',10),(c1,'left',10),(c2,'right',10),
                (tN,'top',5),(t1,'top',5),(tA,'top',5),(b1,'bottom',5)],
            attachControl=[(tN,'left',20,c1),(t1,'right',0,tA),
                (tN,'right',0,t1),(tA,'right',20,c2),(b1,'top',5,tA)],
            attachPosition=[(tA,'left',0,4)],
            attachOppositeControl=[(b1,'left',20,tN),(b1,'right',20,tA)])
        mc.showWindow(w1)
    def exclTSL_update(self):
        '''
        update Exclusion UI lists
        '''
        gc = self.pload()
        mc.textScrollList(self.exclTSL_T, edit=True, removeAll=True)
        mc.textScrollList(self.exclTSL_N, edit=True, removeAll=True)
        if 'type'in gc.exclusions:
            for entry in gc.exclusions['type']:
                src = '->' if entry[2].count('src') else ''
                dst = '->' if entry[2].count('dst') else ''
                mc.textScrollList(self.exclTSL_T, edit=True,
                    append='{0:^4}({2:^52}){1:>3}'.format(dst,src,
                    '{0:>20}.{1:<20}'.format(entry[0],entry[1])))
        if 'name'in gc.exclusions:
            for entry in gc.exclusions['name']:
                src = '->' if entry[2].count('src') else ''
                dst = '->' if entry[2].count('dst') else ''
                mc.textScrollList(self.exclTSL_N, edit=True,
                    append=' {0:^4}({2:^52}){1:>3}'.format(dst,src,
                    '{0:>20}.{1:<20}'.format(entry[0],entry[1])))
        self.psave(gc)
    def exclTSL_del(self,rList):
        gc = self.pload()
        for item in rList:
            mode   = item[0] 
            node   = item[1]
            attr   = item[2]
            target = item[3]
            if (node and attr and target):
                if mode in gc.exclusions:
                    gc.exclusions[mode].remove((node,attr,target))
                else:
                    print ('Cannot remove exclusion which does not exist')
        self.psave(gc)
    